import re
from textual.app import App
from textual.widgets import DataTable, Header, Input, Static
from textual.containers import Container
from datetime import datetime

from .controller import get_tasks, add_task, update_task, TaskStatus
from .db import initialize_db

# TODO: toggle status with 't'
# TODO: add way to filter on other columns
# TODO: safe DB mode
# Nice to Have
# TODO: CLI?
# TODO: config- default category, types, colors
# TODO: dropdowns for status, type, category

column_to_field = {
    1: "text",
    2: "status",
    3: "type",
    4: "due",
    5: "category_name",
}


def remove_rich_tag(text):
    pattern = r"\[[^\]]*\](.*?)\[/\]"
    return re.sub(pattern, r"\1", text)


def get_colored_status(status: str) -> str:
    colors = {
        "zero": "#666666",
        "wip": "#33aa99",
        "blocked": "#cc9900",
        "done": "#009900",
    }
    return f"[{colors[status]}]{status}[/]"


def get_colored_category(category: str) -> str:
    hash_val = sum(ord(c) for c in category)
    hue = hash_val % 360
    color = f"rgb({hue},200,200) on default"
    return f"[{color}]{category}[/]"


class TT(App):
    CSS = """
    #footer {
        dock: bottom;
        max-height: 2;
    }
    #input_bar {
        height: 2;
        border-top: solid green;
        layout: grid;
        grid-size: 2;
        grid-columns: 10 1fr;
        display: none;
    }

    #prompt_label {
        width: 10;
        content-align: left middle;
    }
    
    #input_widget {
        width: 100%;
        margin: 0;
        border: none;
    }

    #status_bar {
      border-top: solid white;
      height: 2;
      margin: 0;
      layout: grid;
      grid-size: 2;
    }

    #left_status {
      height: 1;
      margin: 0;
      padding-left: 1;
    }
    #right_status {
      height: 1;
      margin: 0;
      padding-right: 1;
      text-align: right;
    }
    """

    BINDINGS = [
        # movement
        ("h", "cursor_left", "Left"),
        ("j", "cursor_down", "Down"),
        ("k", "cursor_up", "Up"),
        ("l", "cursor_right", "Right"),
        ("g", "cursor_top", "Top"),
        ("G", "cursor_bottom", "Bottom"),
        # filtering & editing
        ("/", "start_search", "Search"),
        ("f", "start_filter", "Filter"),
        ("c", "start_edit", "Edit Cell"),
        ("escape", "cancel_edit", "Cancel Edit"),
        # other
        ("a", "add_task", "Add"),
        ("d", "delete_task", "Delete"),
        ("q", "quit", "Quit"),
    ]

    def __init__(self):
        super().__init__()
        self.search_query = ""
        self.search_category = ""
        self.saved_cursor_pos = (0, 0)

    def compose(self):
        self.header = Header()
        self.table = DataTable()
        self.input_widget = Input(id="input_widget")
        self.input_label = Static("label", id="prompt_label")
        self.input_bar = Container(id="input_bar")
        self.status_bar = Container(id="status_bar")
        self.left_status = Static("LEFT", id="left_status")
        self.right_status = Static("RIGHT", id="right_status")
        yield self.header
        yield Container(self.table)
        with Container(id="footer"):
            with self.input_bar:
                yield self.input_label
                yield self.input_widget
            with self.status_bar:
                yield self.left_status
                yield self.right_status

    def on_mount(self):
        self.table.add_columns("ID", "Task", "Status", "Type", "Due", "Category")
        self.refresh_tasks(restore_cursor=False)

    def action_cursor_left(self):
        self.table.move_cursor(column=self.table.cursor_column - 1)
        if self.table.cursor_column == 0:
            self.table.cursor_type = "row"

    def action_cursor_right(self):
        self.table.move_cursor(column=self.table.cursor_column + 1)
        if self.table.cursor_column != 0:
            self.table.cursor_type = "cell"

    def action_cursor_up(self):
        self.table.move_cursor(row=self.table.cursor_row - 1)

    def action_cursor_down(self):
        self.table.move_cursor(row=self.table.cursor_row + 1)

    def action_cursor_top(self):
        self.table.move_cursor(row=0)

    def action_cursor_bottom(self):
        self.table.move_cursor(row=self.table.row_count - 1)

    def refresh_tasks(self, *, restore_cursor=True):
        # show table
        self.table.clear()

        tasks = get_tasks(self.search_query, category=self.search_category)
        for task in tasks:
            due_str = task.due.strftime("%Y-%m-%d") if task.due else " - "
            category = get_colored_category(
                task.category.name if task.category else " - "
            )
            status = get_colored_status(task.status)

            self.table.add_row(
                str(task.id),
                task.text,
                status,
                task.type,
                due_str,
                category,
                key=str(task.id),
            )

        # update footer
        if self.search_query:
            self.left_status.update(
                f"{len(tasks)} |{' ' if self.search_category else ''}{self.search_category} matching '{self.search_query}'"
            )
        else:
            self.left_status.update(f"{len(tasks)} | {self.search_category}")

        # restore cursor
        if restore_cursor:
            self.table.move_cursor(
                row=self.saved_cursor_pos[0], column=self.saved_cursor_pos[1]
            )
        else:
            self.table.move_cursor(row=0, column=1)

    def action_delete_task(self):
        if self.table.cursor_column == 0:
            cur_row = self.table.cursor_row
            task_id = int(self.table.get_cell_at((cur_row, 0)))
            update_task(task_id, deleted=True)
            self._save_cursor()
            self.refresh_tasks()

    def action_add_task(self):
        """Add a new task with default values."""
        new_task = add_task(
            text="New Task",
            type="",
            status=TaskStatus.ZERO.value,
            category="main",
        )
        self.refresh_tasks(restore_cursor=False)
        self.move_cursor_to_task(new_task.id)
        self.action_start_edit()

    def move_cursor_to_task(self, task_id):
        # ick, but only way to search table?
        for row in range(self.table.row_count):
            data = self.table.get_row_at(row)
            if data[0] == str(task_id):
                self.table.move_cursor(row=row, column=1)
                break

    # Control of edit bar ####################

    def _show_input(self, mode, start_value):
        self.mode = mode
        self.input_bar.display = True
        self.input_widget.value = start_value
        if mode == "search":
            self.input_label.update("search: ")
        elif mode == "edit":
            self.input_label.update("edit: ")
        elif mode == "filter":
            self.input_label.update("filter: ")
        else:
            raise ValueError(f"unknown mode: {mode}")
        self.set_focus(self.input_widget)

    def _hide_input(self):
        self.input_bar.display = False
        self.set_focus(self.table)

    def action_cancel_edit(self):
        self._hide_input()

    def action_start_search(self):
        self._show_input("search", "")

    def action_start_filter(self):
        self._show_input("filter", self.search_category)

    def _save_cursor(self):
        self.saved_cursor_pos = (self.table.cursor_row, self.table.cursor_column)

    def action_start_edit(self):
        if self.table.cursor_row is None or self.table.cursor_column == 0:
            return

        self._save_cursor()
        current_value = self.table.get_cell_at(
            (self.table.cursor_row, self.table.cursor_column)
        )
        current_value = remove_rich_tag(current_value)
        self._show_input("edit", current_value)

    def on_input_submitted(self, event: Input.Submitted):
        if self.mode == "search":
            self.search_query = event.value
            self.refresh_tasks(restore_cursor=False)
        elif self.mode == "filter":
            self.search_category = event.value
            self.refresh_tasks(restore_cursor=False)
        elif self.mode == "edit":
            self.apply_change(event.value)
        else:
            raise ValueError(f"unknown mode: {self.mode}")
        self._hide_input()

    def apply_change(self, new_value):
        row, col = self.saved_cursor_pos
        task_id = int(self.table.get_cell_at((row, 0)))

        field = column_to_field[col]
        update_data = {}

        if field == "due":
            try:
                update_data["due"] = datetime.strptime(new_value, "%Y-%m-%d")
            except ValueError:
                self.notify("Invalid date format. Use YYYY-MM-DD")
        elif field == "status":
            try:
                TaskStatus(new_value)  # validate status
                update_data["status"] = new_value
            except ValueError:
                self.notify(f"Invalid status. Use: {[s.value for s in TaskStatus]}")
        else:
            update_data[field] = new_value

        try:
            update_task(task_id, **update_data)
            self.refresh_tasks()
        except Exception as e:
            self.notify(f"Error updating task: {str(e)}")
        finally:
            self.action_cancel_edit()


def run():
    initialize_db()
    app = TT()
    app.run()


if __name__ == "__main__":
    run()
